(* pselbutton.ml     Buttons for Psellos apps
 *
 * Copyright (c) 2022 Psellos   http://psellos.com
 *
 * Licensed under the MIT license:
 *     http://opensource.org/licenses/MIT
 *)
open Js_of_ocaml

type state =
    (* State of the button.
     *)
    | Normal   (* Mouse is off somewhere else as far as we know *)
    | OverDown (* Mouse is over the button with mouse button down *)
    | Disabled (* Button is unresponsive *)

type bg =
    | Img of string           (* E.g., Img "img/bpx-blue-308.png" *)
    | Color of CSS.Color.js_t

type t = {
    dombu : Dom_html.buttonElement Js.t; (* The underlying DOM button *)
    mutable state : state;

    (* Foreground functions.
     *)
    normfg : (unit -> unit) ref;
    downfg : (unit -> unit) ref;
    disabfg : (unit -> unit) ref;

    (* Backgrounds (solid color or image)
     *)
    normbg : bg ref;
    downbg : bg ref;
    disabbg : bg ref;
}


let fg_of_state t = function
    | Normal -> t.normfg
    | OverDown -> t.downfg
    | Disabled -> t.disabfg

let bg_of_state t = function
    | Normal -> t.normbg
    | OverDown -> t.downbg
    | Disabled -> t.disabbg


let transit (t : t) newstate =
    let style = t.dombu##.style in
    !(fg_of_state t newstate) ();
    (match !(bg_of_state t newstate) with
    | Img url -> style##.backgroundImage := Pselcss.url_string url
    | Color c -> style##.backgroundColor := (c :> Js.js_string Js.t)
    );
    t.state <- newstate


(* We track mouse up/down/leave to make the background responsive. Could
 * also track enter to be even more responsive.
 *)
let mouse_down t mevent =
    if t.state <> Disabled && t.state <> OverDown then
        transit t OverDown;
    Js._true

let mouse_up t mevent =
    if t.state <> Disabled && t.state <> Normal then
        transit t Normal;
    Js._true

let mouse_out t mevent =
    if t.state <> Disabled && t.state <> Normal then
        transit t Normal;
    Js._true


let create () =
    let clear_color =
        CSS.Color.(js (RGBA_percent (0, 0, 0, 0.0)))
    in
    let button = Dom_html.(
        createButton ~_type: (Js.string "button") document
    )
    in
    let zero = Js.string "0" in
    let style = button##.style in
    style##.position := Js.string "absolute";
    style##.padding := zero;
    style##.margin := zero;
    style##.borderWidth := zero;
    style##.outlineWidth := zero;
    style##.backgroundColor := (clear_color :> Js.js_string Js.t);

    let nofun = (fun () -> ())in
    let res =
    {
        dombu = button;
        state = Normal;
        normfg = ref nofun;
        downfg = ref nofun;
        disabfg = ref nofun;
        normbg = ref (Color clear_color);
        downbg = ref (Color clear_color);
        disabbg = ref (Color clear_color);
    }
    in
    button##.onmousedown := Dom.handler (mouse_down res);
    button##.onmouseup := Dom.handler (mouse_up res);
    button##.onmouseout := Dom.handler (mouse_out res);
    res

let element t =
    t.dombu

let width (t : t) =
    (* Width in pixels. To make sure this works, must use set_width to
     * set the width.
     *)
    Pselcss.px_length t.dombu##.style##.width

let set_width (t : t) width =
    t.dombu##.style##.width := Pselcss.length_px width

let height (t : t) =
    (* Height in pixels. To make sure this works, must use set_height to
     * set the height.
     *)
    Pselcss.px_length t.dombu##.style##.height

let set_height (t : t) height =
    t.dombu##.style##.height := Pselcss.length_px height

let set_inner (t : t) children =
    let rec clear () =
        match Js.Opt.to_option t.dombu##.firstChild with
        | None -> ()
        | Some n -> Dom.removeChild t.dombu n; clear ()
    in
    clear ();
    List.iter (Dom.appendChild t.dombu) children

let set_look t state fgfun bg =
    fg_of_state t state := fgfun;
    bg_of_state t state := bg;
    if state = t.state then
        transit t state

let set_topleft (t : t) x y =
    let style = t.dombu##.style in
    style##.left := Pselcss.length_px x;
    style##.top := Pselcss.length_px y

let set_onclick (t : t) fn_opt =
    match fn_opt with
    | None -> t.dombu##.onclick := Dom.no_handler
    | Some fn -> t.dombu##.onclick := Dom.handler (fun ev -> Js.bool (fn ev))

let set_disabled t disabled =
    if (t.state = Disabled) <> disabled then (
        t.dombu##.disabled := Js.bool disabled;
        let state' = if disabled then Disabled else Normal in
        transit t state'
    )
