(* s24defs.ml     Common definitions for Slide24 example
 *
 * Copyright (c) 2020 Psellos   http://psellos.com
 *
 * Licensed under the MIT license:
 *     http://opensource.org/licenses/MIT
 *)

let fround f =
    floor (f +. 0.5)

let round f =
    truncate (f +. 0.5)

let rec range i j =
    if i > j then [] else i :: range (i + 1) j

let rec drange i j =
    if i < j then [] else i :: drange (i - 1) j

let udrange i j =
    if i <= j then range i j else drange i j

let array_findi p arr =
    let n = Array.length arr in
    let rec loop i =
        if i = n then raise Not_found
        else if p arr.(i) then i
        else loop (i + 1)
    in
    loop 0

(* We want to support some number of shuffle buttons. Each one has a
 * label, a number of steps to shuffle, and a boolean saying whether the
 * shuffling should start from the solved position (true) or from the
 * current position (false).
 *)
type shufd = Shufd of string * int * bool


(* A configuration of tiles is represented by an array C containing a
 * permutation of 0..24, with 0 representing the empty spot. C.(0) tells
 * which tile is at the upper left, C.(1) is the tile just to the right
 * of this.
 *)
type config = int array

module ConfigSet =
    Set.Make (struct type t = config let compare = compare end)


let shuffle_small config n =
    (* Shuffle by a small amount. For small shuffles it's worth making
     * sure that the shuffle makes progress rather than circling back to
     * a previous configuration.
     *)
    let moves x0 =
        let res = if x0 >= 5 then [x0 - 5] else [] in
        let res' = if x0 < 20 then (x0 + 5) :: res else res in
        let res'' = if x0 mod 5 <> 0 then (x0 - 1) :: res' else res' in
        if x0 mod 5 <> 4 then (x0 + 1) :: res'' else res''
    in
    let apply emptyx newx =
        let config' = Array.copy config in
        config'.(emptyx) <- config'.(newx);
        config'.(newx) <- 0;
        (newx, config')
    in
    let rec i_shuffle visited emptyx k =
        if k > 0 then
            let newxs = moves emptyx in
            let newconfigs = List.map (apply emptyx) newxs in
            let goodconfigs =
                List.filter
                    (fun (_, c) -> not (ConfigSet.mem c visited))
                    newconfigs
            in
            let configs =
                if goodconfigs = [] then newconfigs else goodconfigs
            in
            let (emptyx', c) =
                List.nth configs (Random.int (List.length configs))
            in
            Array.blit c 0 config 0 (Array.length config);
            i_shuffle (ConfigSet.add c visited) emptyx' (k - 1)
    in
    i_shuffle ConfigSet.empty (array_findi ((=) 0) config) n


let shuffle config n =
    (* Shuffle the board by moving the empty spot around randomly n
     * times. It's an inefficient way to shuffle, but you can get
     * different amounts of shuffling by adjusting n.
     *)
    if n < 20 then
        shuffle_small config n
    else
        let rec ishuffle emptyx k =
            if k > 0 then
                let mtr = emptyx / 5 in
                let mtc = emptyx mod 5 in
                let (mtr', mtc') =
                    if Random.bool () then
                        (mtr, mtc + Random.int 2 * 2 - 1)
                    else
                        (mtr + Random.int 2 * 2 - 1, mtc)
                in
                if mtr' >= 0 && mtr' < 5 && mtc' >= 0 && mtc' < 5 then (
                    let emptyx' = mtr' * 5 + mtc' in
                    config.(emptyx) <- config.(emptyx');
                    config.(emptyx') <- 0;
                    ishuffle emptyx' (k - 1)
                ) else (
                    ishuffle emptyx k
                )
        in
        let empty_index = array_findi ((=) 0) config in
        ishuffle empty_index n


let coalesce_blocs configs =
    (* Coalesce sequences of individual moves into bloc moves.
     *
     * The essence is to collapse sequences of configs with equal deltas
     * down to just the last config in the sequence.
     *)
    let delta a b =
        (* Determine direction of tile movement in the transition from
         * config a to config b. One of (0, 0), (+-1, 0), (0, +-1)
         *)
        let emptya = array_findi ((=) 0) a in
        let emptyb = array_findi ((=) 0) b in
        (emptya / 5 - emptyb / 5, emptya mod 5 - emptyb mod 5)
    in
    let rec mkdeltas accum configs =
        match configs with
        | [] | [_] -> (List.rev accum)
        | a :: (b :: _ as rest) -> mkdeltas (delta a b :: accum) rest
    in
    let rec uniq accum configs deltas =
        (* [Configs and deltas have the same length by construction.]
         *)
        match configs, deltas with
        | [], _ | _, [] -> List.rev accum
        | [c], _ -> List.rev (c :: accum)
        | _ :: _ :: _, [_] -> List.rev_append accum configs (* Not possible *)
        | ca :: (cb :: _ as crest), da :: (db :: _ as drest) ->
            let accum' =
                if da = db then accum else ca :: accum
            in
            uniq accum' crest drest
    in
    let deltas =
        match configs with
        | [] -> []
        | h :: _ -> mkdeltas [] (h :: configs)
    in
    uniq [] configs deltas
