(* s24look.ml     Visual look of Slide24
 *
 * Copyright (c) 2022 Psellos   http://psellos.com
 *
 * Licensed under the MIT license:
 *     http://opensource.org/licenses/MIT
 *)
open Js_of_ocaml

type rect = float * float * float * float      (* x, y, w, h *)
type point = float * float                     (* x, y *)
type size = float * float                      (* w, h *)

type bdcolor = float * float * float * float

let bdwhite = (1.0, 1.0, 1.0, 1.0)
let bdred = (1.0, 0.0, 0.0, 1.0)

let clear = Pselcss.color_bdcolor (0.0, 0.0, 0.0, 0.0)
let white = Pselcss.color_bdcolor bdwhite
let red = Pselcss.color_bdcolor bdred
let dkwhite = Pselcss.color_bdcolor (0.8, 0.8, 0.8, 1.0)
let dkred = Pselcss.color_bdcolor (0.8, 0.0, 0.0, 1.0)

(* Size of shuffle and solve buttons.
 *)
let shuffle_size = (125.0, 44.0)

(* The look of a font: (style, weight, family). The style and weight can
 * be "" to indicate the normal value.
 *
 * Family should be of the form "a, b" where b is a generic font.
 *
 * Some good styles are: normal, italic, oblique
 * Some good weights are: normal, bold
 * Some generic families are: serif, sans-serif, monospace
 *)
type font_look = string * string * string


let font_spec (fstyle, fweight, family) fsize =
    (* Calculate a font specification in the form accepted by CSS.
     *)
    let (specific_fam, generic_fam) =
        match String.index family ',' with
        | n ->
            let flen = String.length family in
            (String.sub family 0 n, String.sub family (n + 1) (flen - n - 1))
        | exception Not_found -> (family, "")
    in
    let fweight' = if fweight <> "" then " " ^ fweight else "" in
    let fstyle' = if fstyle <> "" then " " ^ fstyle else "" in
    let generic_fam' = if generic_fam <> "" then ", " ^ generic_fam else "" in
    let spec =
        Printf.sprintf "%s%s %dpx \"%s\"%s"
            fstyle' fweight' (truncate fsize) specific_fam generic_fam'
    in
    Js.string spec


let font_size flook teststr width initial =
    (* Figure out what font size to use (in pixel units) so that the
     * given test string fits in the given width.
     *
     * Assume that the result is between initial - 10 and initial + 10
     * (inclusive).
     *
     * Some (possibly all) browsers do not deal in fractional pixel
     * sizes for fonts. So we deal only in whole numbers.
     *)
     let js_teststr = Js.string teststr in
     let canv = Dom_html.createCanvas Dom_html.document in
     let ctx2d = canv##getContext Dom_html._2d_ in
     let rec loop iters lo hi =
         (* Invariants
          * lo < hi
          * text widths for sizes <= lo are <= width
          * text widths for sizes >= hi are > width
          *)
        if iters > 6 || lo >= hi -. 1.0 then (
            lo
        ) else (
            let mid = floor ((lo +. hi) /. 2.0) in
            ctx2d##.font := font_spec flook mid;
            let text_width = (ctx2d##measureText js_teststr)##.width in
            if text_width <= width then
                loop (iters + 1) mid hi
            else
                loop (iters + 1) lo mid
        )
    in
    let initial' = floor initial in
    loop 0 (initial' -. 10.0) (initial' +. 10.0)
